@@ -4,12 +4,12 @@ where
 -- version number is major.minor.patchlvl; don't change the format of the
 -- `versnum' line as it is `grep'ed for by a Makefile
 --
-idstr      = "$Id: BaseVersion.hs,v 1.32 2001/04/29 12:23:46 chak Exp $"
+idstr      = "$Id: BaseVersion.hs,v 1.43 2003/10/19 10:45:27 chak Exp $"
 name       = "Compiler Toolkit"
-versnum    = "0.22.3"
-date	   = "21 Apr 2001"
+versnum    = "0.26.0"
+date	   = "19 Oct 2003"
 version    = name ++ ", version " ++ versnum ++ ", " ++ date
-copyright  = "Copyright (c) [1995..2001] Manuel M. T. Chakravarty"
+copyright  = "Copyright (c) [1995..2003] Manuel M T Chakravarty"
 disclaimer = "This software is distributed under the \
 	     \terms of the GNU Public Licence.\n\
 	     \NO WARRANTY WHATSOEVER IS PROVIDED. \
