@@ -3,7 +3,7 @@
 --  Author : Manuel M. T. Chakravarty
 --  Created: 6 March 99
 --
---  Version $Revision: 1.16 $ from $Date: 2001/08/24 14:42:04 $
+--  Version $Revision: 1.20 $ from $Date: 2003/10/19 10:46:09 $
 --
 --  Copyright (c) [1999..2001] Manuel M. T. Chakravarty
 --
@@ -28,7 +28,8 @@
 --  We assume that the input already went through cpp.  Thus, we do not handle 
 --  comments and preprocessor directives here.  The lexer recognizes all tokens
 --  of ANCI C except those occuring only in function bodies.  It supports the
---  C99 `restrict' extension: <http://www.lysator.liu.se/c/restrict.html>.
+--  C99 `restrict' extension: <http://www.lysator.liu.se/c/restrict.html> as
+--  well as inline functions.
 --
 --  Comments:
 --
@@ -42,7 +43,7 @@
 --    identifiers). 
 --
 --  * We also recognize GNU C `__attribute__', `__extension__', `__const', 
---    `__const__', `__restrict', and `__restrict__'.
+--    `__const__', `__inline', `__inline__', `__restrict', and `__restrict__'.
 --
 --  * Any line starting with `#pragma' is ignored.
 --
@@ -131,6 +132,9 @@ data CToken = CTokLParen   Position		-- `('
 	    | CTokLBrace   Position		-- `{'
 	    | CTokRBrace   Position		-- `}'
 	    | CTokEllipsis Position		-- `...'
+	    | CTokAlignof  Position		-- `alignof' 
+						-- (or `__alignof', 
+						-- `__alignof__')
 	    | CTokChar     Position		-- `char'
 	    | CTokConst    Position		-- `const' 
 						-- (or `__const', `__const__')
@@ -138,6 +142,9 @@ data CToken = CTokLParen   Position		-- `('
 	    | CTokEnum     Position		-- `enum'
 	    | CTokExtern   Position		-- `extern'
 	    | CTokFloat    Position		-- `float'
+	    | CTokInline   Position		-- `inline'
+						-- (or `__inline', 
+						-- `__inline__')
 	    | CTokInt      Position		-- `int'
 	    | CTokLong     Position		-- `long'
 	    | CTokRestrict Position		-- `restrict'
@@ -215,6 +222,7 @@ instance Pos CToken where
   posOf (CTokLBrace   pos  ) = pos
   posOf (CTokRBrace   pos  ) = pos
   posOf (CTokEllipsis pos  ) = pos
+  posOf (CTokAlignof  pos  ) = pos
   posOf (CTokChar     pos  ) = pos
   posOf (CTokConst    pos  ) = pos
   posOf (CTokDouble   pos  ) = pos
@@ -289,12 +297,14 @@ instance Show CToken where
   showsPrec _ (CTokLBrace   _  ) = showString "{"
   showsPrec _ (CTokRBrace   _  ) = showString "}"
   showsPrec _ (CTokEllipsis _  ) = showString "..."
+  showsPrec _ (CTokAlignof  _  ) = showString "alignof"
   showsPrec _ (CTokChar     _  ) = showString "char"
   showsPrec _ (CTokConst    _  ) = showString "const"
   showsPrec _ (CTokDouble   _  ) = showString "double"
   showsPrec _ (CTokEnum     _  ) = showString "enum"
   showsPrec _ (CTokExtern   _  ) = showString "extern"
   showsPrec _ (CTokFloat    _  ) = showString "float"
+  showsPrec _ (CTokInline   _  ) = showString "inline"
   showsPrec _ (CTokInt      _  ) = showString "int"
   showsPrec _ (CTokLong     _  ) = showString "long"
   showsPrec _ (CTokRestrict _  ) = showString "restrict"
@@ -365,12 +375,14 @@ instance Tag CToken where
   tag (CTokLBrace   _  ) = 44
   tag (CTokRBrace   _  ) = 45
   tag (CTokEllipsis _  ) = 46
+  tag (CTokAlignof  _  ) = 74
   tag (CTokChar     _  ) = 47
   tag (CTokConst    _  ) = 48
   tag (CTokDouble   _  ) = 49
   tag (CTokEnum     _  ) = 50
   tag (CTokExtern   _  ) = 51
   tag (CTokFloat    _  ) = 52
+  tag (CTokInline   _  ) = 75
   tag (CTokInt      _  ) = 53
   tag (CTokLong     _  ) = 54
   tag (CTokRestrict _  ) = 55
@@ -392,6 +404,8 @@ instance Tag CToken where
   tag (CTokTypeName _ _) = 71
   tag (CTokGnuC GnuCAttrTok _) = 72
   tag (CTokGnuC GnuCExtTok  _) = 73
+  -- current max is 75 (see `CTokInline')
+
 
 -- local state instantiation
 -- -------------------------
@@ -443,11 +457,12 @@ whitespace  =      (char ' ' `lexaction` \_ _ -> Nothing)
 
 -- #line directive (K&R A12.6)
 --
--- * allows further numbers to follow the file name a la GCC
+-- * allows further ints after the file name a la GCC; as the GCC CPP docu
+--   doesn't say how many ints there can be, we allow an unbound number
 --
 linedir :: CLexer
 linedir  = char '#' +> ppwhite +> int +> ppwhite +> (fname +> ppwhite)`quest`
-	   ((int +> ppwhite +> (int +> ppwhite)`quest`epsilon)`quest` 
+	   ((int +> ppwhite)`star` 
 	    char '\n')
 	   `lexmeta` \str pos ns -> (Nothing, adjustPos str pos, ns, Nothing)
 	   where
@@ -468,7 +483,7 @@ linedir  = char '#' +> ppwhite +> int +> ppwhite +> (fname +> ppwhite)`quest`
 		 --
 		 dropWhite = dropWhile (\c -> c == ' ' || c == '\t')
 
--- #pragma directive (K&R A12.?)
+-- #pragma directive (K&R A12.8)
 --
 -- * we simply ignore any #pragma (but take care to update the position
 --   information)
@@ -493,6 +508,9 @@ identOrKW  =
   letter +> (letter >|< digit)`star` epsilon
   `lexactionName` \cs pos name -> (idkwtok $!pos) cs name
   where
+    idkwtok pos "alignof"       _    = CTokAlignof  pos
+    idkwtok pos "__alignof"     _    = CTokAlignof  pos
+    idkwtok pos "__alignof__"   _    = CTokAlignof  pos
     idkwtok pos "char"          _    = CTokChar     pos
     idkwtok pos "const"		_    = CTokConst    pos
     idkwtok pos "__const"	_    = CTokConst    pos
@@ -501,6 +519,9 @@ identOrKW  =
     idkwtok pos "enum"		_    = CTokEnum     pos
     idkwtok pos "extern"	_    = CTokExtern   pos
     idkwtok pos "float"		_    = CTokFloat    pos
+    idkwtok pos "inline"	_    = CTokInline   pos
+    idkwtok pos "__inline"	_    = CTokInline   pos
+    idkwtok pos "__inline__"	_    = CTokInline   pos
     idkwtok pos "int"		_    = CTokInt      pos
     idkwtok pos "long"		_    = CTokLong     pos
     idkwtok pos "restrict"	_    = CTokRestrict pos
