@@ -1,11 +1,11 @@
 --  C -> Haskell Compiler: Parser for C Header Files
 --
---  Author : Manuel M. T. Chakravarty
+--  Author : Manuel M T Chakravarty
 --  Created: 7 March 99
 --
---  Version $Revision: 1.15 $ from $Date: 2001/08/24 14:42:04 $
+--  Version $Revision: 1.20 $ from $Date: 2003/02/12 09:41:02 $
 --
---  Copyright (c) [1999..2001] Manuel M. T. Chakravarty
+--  Copyright (c) [1999..2002] Manuel M T Chakravarty
 --
 --  This file is free software; you can redistribute it and/or modify
 --  it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@
 --  The parser recognizes all of ANCI C except function bodies.  The parser
 --  combinators follow K&R Appendix A, but we make use of the richer grammar
 --  constructs provided by `Parsers'.  It supports the C99 `restrict'
---  extension.
+--  extension and `inline'.
 --
 --  Comments:
 --
@@ -48,12 +48,16 @@
 --  * We also recognize GNU C `__attribute__' annotations (however, they are
 --    not entered into the structure tree, but ignored).  More specifically, 
 --
---      __attribute__ (( <attr> ))
+--      '__attribute__' '(' '(' attr ')' ')'
 --
---    may occur as a prefix and/or suffix of a `declarator', where <attr>
---    is either just an identifier or an identifier followed by a
---    comma-separated list of arguments, which may be numbers, strings, and
---    identifiers. 
+--    may occur after declarator specifiers or after a declarator itself (only
+--    meaningful if it is a typedef), where `attr' is either just an 
+--    identifier or an identifier followed by a comma-separated list of
+--    constant expressions as follows:
+--
+--      attr  -> id ['(' const_1 ',' ... ',' const_n ')']
+--	       | 'const'
+--	const -> <constant expression>
 --
 --  * We also recognize GNU C `__extension__' annotations (however, they are
 --    not entered into the structure tree, but ignored).  More specifically, 
@@ -281,6 +285,7 @@ parseCHeader pos tokens  =
     -- typedef-name by a `typedef' declaration
     --
     parseCExtDeclList      :: [CToken] -> CST s [CExtDecl]
+    parseCExtDeclList []    = return []
     parseCExtDeclList toks  = 
       do
 	nameSupply <- getNameSupply
@@ -336,10 +341,17 @@ parseCExtDecl  = parseCDecl
 
 -- parse C declaration (K&R A8)
 --
+-- * We allow the GNU C extension keyword before a declaration and GNU C
+--   attribute annotations after declaration specifiers, but they are not
+--   entered into the structure tree.
+--
 parseCDecl :: CParser CDecl
 parseCDecl  = 
-  ctoken_ (CTokGnuC GnuCExtTok) `opt` ()        -- ignore GCC's __extension__
-  -*> list parseCDeclSpec *> seplist comma_ parseCInitDecl *-> semic_
+  list (
+      ctoken_ (CTokGnuC GnuCExtTok) `opt` ()      -- ignore GCC's __extension__
+  -*> parseCDeclSpec 
+  *-> optMaybe parseGnuCAttr			  -- ignore GCC's __attribute__
+  )*> seplist comma_ parseCInitDecl *-> semic_
   `actionAttrs`
     (\(specs, declrs) -> 
       head (map posOf specs ++ map (posOf . fst) declrs ++ [nopos])) $
@@ -348,6 +360,24 @@ parseCDecl  =
 			   in
 			   CDecl specs declrs' at
 
+-- parse GNU C attribute annotation (junking the result)
+--
+parseGnuCAttr :: CParser ()
+parseGnuCAttr  = 
+     ctoken_ (CTokGnuC GnuCAttrTok) 
+  *> ctoken_ CTokLParen *> ctoken_ CTokLParen
+  *> parseAttr
+  *> ctoken_ CTokRParen *> ctoken_ CTokRParen
+  `action`
+    const ()
+  where
+    parseAttr    =     cid *> optMaybe (   ctoken_ CTokLParen 
+					*> seplist comma_ parseAttrArg
+					*> ctoken_ CTokRParen)
+			`action` const ()
+                    <|> ctoken_ CTokConst `action` const ()
+    parseAttrArg =  parseCConstExpr `action` const ()
+
 -- parse C declaration specifier (K&R A8)
 --
 parseCDeclSpec :: CParser CDeclSpec
@@ -399,12 +429,13 @@ parseCTypeSpec  =     CTokVoid     `tokenToAttrs` CVoidType
 
 -- parse C type qualifier (K&R A8.2)
 --
--- * plus `restrict' from C99
+-- * plus `restrict' from C99 and `inline'
 --
 parseCTypeQual :: CParser CTypeQual
 parseCTypeQual  =     CTokConst    `tokenToAttrs` CConstQual
 		  <|> CTokVolatile `tokenToAttrs` CVolatQual
 		  <|> CTokRestrict `tokenToAttrs` CRestrQual
+		  <|> CTokInline   `tokenToAttrs` CInlinQual
 
 -- parse C structure of union declaration (K&R A8.3)
 --
@@ -431,9 +462,13 @@ parseCStructUnion  =
 
 -- parse C structure declaration (K&R A8.3)
 --
+-- * We allow the GNU C extension keyword before a declaration, but it is
+--   not entered into the structure tree.
+--
 parseCStructDecl :: CParser CDecl
 parseCStructDecl  =
-  list parseCSpecQual *> seplist comma_ parseCStructDeclr *-> semic_
+      ctoken_ (CTokGnuC GnuCExtTok) `opt` ()      -- ignore GCC's __extension__
+  -*> list parseCSpecQual *> seplist comma_ parseCStructDeclr *-> semic_
   `actionAttrs`
     (\(specs, declrs) -> 
       head (map posOf specs ++ map declrsPos declrs ++ [nopos])) $
@@ -493,16 +528,15 @@ parseCEnum  =
 -- * the parser for `optPointer' also returns a function; it is applied last,
 --   because indirection has the least precedence
 --
--- * We allow GNU C attribute annotations as prefix and suffix of declarators, 
+-- * We allow GNU C attribute annotations at the end of a declerator,
 --   but they are not entered into the structure tree.
 --
 parseCDeclr :: CParser CDeclr
 parseCDeclr  =
-      optMaybe parseGnuCAttr
-  -*> (pointer `opt` id)
+      (pointer `opt` id)
   *>  base
   *>  many (flip (.)) id (arrayType <|> newStyleFun <|> oldStyleFun)
-  *-> optMaybe parseGnuCAttr
+  *-> optMaybe parseGnuCAttr			  -- ignore GCC's __attribute__
   `action`
     \((ptr, base), declrTrans) -> ptr . declrTrans $ base 
   where
@@ -544,26 +578,6 @@ parseCDeclr  =
 		  id $
 		  \_ at -> (\declr -> CFunDeclr declr [] False at)
 
--- parse GNU C attribute annotation (junking the result)
---
-parseGnuCAttr :: CParser ()
-parseGnuCAttr  = 
-     ctoken_ (CTokGnuC GnuCAttrTok) 
-  *> ctoken_ CTokLParen *> ctoken_ CTokLParen
-  *> parseAttr
-  *> ctoken_ CTokRParen *> ctoken_ CTokRParen
-  `action`
-    const ()
-  where
-    parseAttr    =     cid *> optMaybe (   ctoken_ CTokLParen 
-					*> seplist comma_ parseAttrArg
-					*> ctoken_ CTokRParen)
-			`action` const ()
-                    <|> ctoken_ CTokConst `action` const ()
-    parseAttrArg =      cid  `action` const ()
-		    <|> cint `action` const ()
-		    <|> cstr `action` const ()
-
 -- parse C parameter type list (K&R A8.6.3)
 --
 parseCParmTypeList :: CParser ([CDecl], Bool)
@@ -725,6 +739,8 @@ parseCPostfixExpr  =
 
 -- parse C unary expression (K&R A7.4)
 --
+-- * GNU extension: `alignof'
+--
 parseCUnaryExpr :: CParser CExpr
 parseCUnaryExpr  = 
       parseCPostfixExpr
@@ -749,6 +765,15 @@ parseCUnaryExpr  =
       *-> ctoken_ CTokLParen *> parseCTypeName *-> ctoken_ CTokRParen
       `action`
         (\(at, tname) -> CSizeofType tname at)
+  <|>
+      ctokenAttrs CTokAlignof *> parseCUnaryExpr
+      `action`
+        (\(at, expr) -> CAlignofExpr expr at)
+  <|>
+         ctokenAttrs CTokAlignof 
+      *-> ctoken_ CTokLParen *> parseCTypeName *-> ctoken_ CTokRParen
+      `action`
+        (\(at, tname) -> CAlignofType tname at)
   where
     parseCUnaryOp =     ctoken CTokAmper  `opAction` CAdrOp
 		    <|> ctoken CTokStar   `opAction` CIndOp
@@ -953,7 +978,7 @@ parseCConst  =     (cint   `actionAttrs` snd $ CIntConst . fst)
 -- main parsing routine
 -- --------------------
 
--- parse the cpp-preprocessed C header contained contained in the given string 
+-- parse the cpp-preprocessed C header contained in the given string 
 -- (EXPORTED)
 --
 -- * the given position is attributed to the first character in the string
