@@ -1,11 +1,11 @@
 --  C->Haskell Compiler: CHS file abstraction
 --
---  Author : Manuel M. T. Chakravarty
+--  Author : Manuel M T Chakravarty
 --  Created: 16 August 99
 --
---  Version $Revision: 1.16 $ from $Date: 2001/06/20 09:25:13 $
+--  Version $Revision: 1.24 $ from $Date: 2003/10/19 10:46:10 $
 --
---  Copyright (c) [1999..2001] Manuel M. T. Chakravarty
+--  Copyright (c) [1999..2003] Manuel M T Chakravarty
 --
 --  This file is free software; you can redistribute it and/or modify
 --  it under the terms of the GNU General Public License as published by
@@ -54,14 +54,18 @@
 --	      | `type' ident
 --	      | `sizeof' ident
 --	      | `enum' idalias trans [`with' prefix] [deriving]
---	      | `call' [`fun'] [`unsafe'] idalias
+--	      | `call' [`pure'] [`unsafe'] idalias
+--	      | `fun' [`pure'] [`unsafe'] idalias parms
 --	      | `get' apath
 --	      | `set' apath
 --	      | `pointer' ['*'] idalias ptrkind
---  ctxt     -> [`header' `=' string] [`lib' `=' string] [prefix]
---  idalias  -> ident [`as' ident]
+--	      | `class' [ident `=>'] ident ident
+--  ctxt     -> [`lib' `=' string] [prefix]
+--  idalias  -> ident [`as' (ident | `^')]
 --  prefix   -> `prefix' `=' string
 --  deriving -> `deriving' `(' ident_1 `,' ... `,' ident_n `)'
+--  parms    -> [verbhs `=>'] `{' parm_1 `,' ... `,' parm_n `}' `->' parm
+--  parm     -> [ident_1 [`*' | `-']] verbhs [`&'] [ident_2 [`*' | `-']]
 --  apath    -> ident
 --	      | `*' apath
 --	      | apath `.' ident
@@ -77,26 +81,33 @@
 --  Remark: Optional Haskell names are normalised during structure tree
 --	    construction, ie, associations that associated a name with itself
 --	    are removed.  (They don't carry semantic content, and make some
---	    test more complicated.)
+--	    tests more complicated.)
 --
 --- TODO ----------------------------------------------------------------------
 --
 
-module CHS (CHSModule(..), CHSFrag(..), CHSHook(..), CHSTrans(..),
-	    CHSAccess(..), CHSAPath(..), CHSPtrType(..),
-	    loadCHS, dumpCHS, hssuffix, chssuffix, loadCHI, dumpCHI, chisuffix)
+module CHS (CHSModule(..), CHSFrag(..), CHSHook(..), CHSTrans(..), CHSParm(..),
+	    CHSArg(..), CHSAccess(..), CHSAPath(..), CHSPtrType(..),
+	    loadCHS, dumpCHS, hssuffix, chssuffix, loadCHI, dumpCHI,
+	    chisuffix, showCHSParm)
 where 
 
+-- standard libraries
+import Char	 (isSpace, toUpper, toLower)
 import List	 (intersperse)
 import Monad	 (when)
 
-import Common    (Position, Pos(posOf), nopos)
+-- Compiler Toolkit
+import Common    (Position, Pos(posOf), nopos, isBuiltinPos)
 import Errors	 (interr)
-import Idents    (Ident, identToLexeme)
+import Idents    (Ident, identToLexeme, onlyPosIdent)
 
+-- C->Haskell
 import C2HSState (CST, doesFileExistCIO, readFileCIO, writeFileCIO, getId, 
 		  getSwitch, chiPathSB, catchExc, throwExc, raiseError, 
 		  fatal, errorsPresent, showErrors, Traces(..), putTraceStr) 
+
+-- friends
 import CHSLexer  (CHSToken(..), lexCHS)
 
 
@@ -109,8 +120,38 @@ data CHSModule = CHSModule [CHSFrag]
 
 -- a CHS code fragament (EXPORTED)
 --
-data CHSFrag = CHSVerb String
-	     | CHSHook CHSHook
+-- * `CHSVerb' fragments are present throughout the compilation and finally
+--   they are the only type of fragment (describing the generated Haskell
+--   code)
+--
+-- * `CHSHook' are binding hooks, which are being replaced by Haskell code by
+--   `GenBind.expandHooks' 
+--
+-- * `CHSCPP' and `CHSC' are fragements of C code that are being removed when
+--   generating the custom C header in `GenHeader.genHeader'
+--
+-- * `CHSCond' are strutured conditionals that are being generated by
+--   `GenHeader.genHeader' from conditional CPP directives (`CHSCPP')
+--
+data CHSFrag = CHSVerb String			-- Haskell code
+		       Position
+	     | CHSHook CHSHook			-- binding hook
+	     | CHSCPP  String			-- pre-processor directive
+		       Position
+	     | CHSC    String			-- C code
+		       Position
+	     | CHSCond [(Ident,			-- C variable repr. condition
+		         [CHSFrag])]		-- then/elif branches
+		       (Maybe [CHSFrag])	-- else branch
+
+instance Pos CHSFrag where
+  posOf (CHSVerb _ pos ) = pos
+  posOf (CHSHook hook  ) = posOf hook
+  posOf (CHSCPP  _ pos ) = pos
+  posOf (CHSC    _ pos ) = pos
+  posOf (CHSCond alts _) = case alts of
+			     (_, frag:_):_ -> posOf frag
+			     _             -> nopos
 
 -- a CHS binding hook (EXPORTED)
 --
@@ -118,8 +159,7 @@ data CHSHook = CHSImport  Bool			-- qualified?
 			  Ident			-- module name
 			  String		-- content of .chi file
 			  Position
-	     | CHSContext (Maybe String)	-- header name
-			  (Maybe String)	-- library name
+	     | CHSContext (Maybe String)	-- library name
 			  (Maybe String)	-- prefix
 			  Position
 	     | CHSType    Ident			-- C type
@@ -137,6 +177,14 @@ data CHSHook = CHSImport  Bool			-- qualified?
 			  Ident			-- C function
 			  (Maybe Ident)		-- Haskell name
 			  Position
+	     | CHSFun     Bool			-- is a pure function?
+			  Bool			-- is unsafe?
+			  Ident			-- C function
+			  (Maybe Ident)		-- Haskell name
+			  (Maybe String)	-- type context
+			  [CHSParm]		-- argument marshalling
+			  CHSParm		-- result marshalling
+			  Position
 	     | CHSField   CHSAccess		-- access type
 			  CHSAPath		-- access path
 			  Position 
@@ -147,16 +195,22 @@ data CHSHook = CHSImport  Bool			-- qualified?
 			  Bool			-- create new type?
 			  (Maybe Ident)		-- Haskell type pointed to
 			  Position
+             | CHSClass   (Maybe Ident)		-- superclass
+			  Ident			-- class name
+			  Ident			-- name of pointer type
+			  Position
 
 instance Pos CHSHook where
-  posOf (CHSImport  _ _ _       pos) = pos
-  posOf (CHSContext _ _ _       pos) = pos
-  posOf (CHSType    _           pos) = pos
-  posOf (CHSSizeof  _           pos) = pos
-  posOf (CHSEnum    _ _ _ _ _   pos) = pos
-  posOf (CHSCall    _ _ _ _     pos) = pos
-  posOf (CHSField   _ _         pos) = pos
-  posOf (CHSPointer _ _ _ _ _ _ pos) = pos
+  posOf (CHSImport  _ _ _         pos) = pos
+  posOf (CHSContext _ _           pos) = pos
+  posOf (CHSType    _             pos) = pos
+  posOf (CHSSizeof  _             pos) = pos
+  posOf (CHSEnum    _ _ _ _ _     pos) = pos
+  posOf (CHSCall    _ _ _ _       pos) = pos
+  posOf (CHSFun     _ _ _ _ _ _ _ pos) = pos
+  posOf (CHSField   _ _           pos) = pos
+  posOf (CHSPointer _ _ _ _ _ _   pos) = pos
+  posOf (CHSClass   _ _ _	  pos) = pos
 
 -- two hooks are equal if they have the same Haskell name and reference the
 -- same C object 
@@ -164,8 +218,8 @@ instance Pos CHSHook where
 instance Eq CHSHook where
   (CHSImport qual1 ide1 _      _) == (CHSImport qual2 ide2 _      _) =    
     qual1 == qual2 && ide1 == ide2
-  (CHSContext h1 olib1 opref1  _) == (CHSContext h2 olib2 opref2  _) =    
-    h1 == h2 && olib1 == olib1 && opref1 == opref2
+  (CHSContext olib1 opref1  _   ) == (CHSContext olib2 opref2  _   ) =    
+    olib1 == olib1 && opref1 == opref2
   (CHSType ide1                _) == (CHSType ide2                _) = 
     ide1 == ide2
   (CHSSizeof ide1              _) == (CHSSizeof ide2              _) = 
@@ -174,11 +228,16 @@ instance Eq CHSHook where
     oalias1 == oalias2 && ide1 == ide2
   (CHSCall _ _ ide1 oalias1    _) == (CHSCall _ _ ide2 oalias2    _) = 
     oalias1 == oalias2 && ide1 == ide2
+  (CHSFun  _ _ ide1 oalias1 _ _ _ _) 
+			          == (CHSFun _ _ ide2 oalias2 _ _ _ _) = 
+    oalias1 == oalias2 && ide1 == ide2
   (CHSField acc1 path1         _) == (CHSField acc2 path2         _) =    
     acc1 == acc2 && path1 == path2
   (CHSPointer _ ide1 oalias1 _ _ _ _) 
 			          == (CHSPointer _ ide2 oalias2 _ _ _ _) =
     ide1 == ide2 && oalias1 == oalias2
+  (CHSClass _ ide1 _           _) == (CHSClass _ ide2 _           _) =
+    ide1 == ide2
   _                               == _                          = False
 
 -- translation table (EXPORTED)
@@ -186,6 +245,24 @@ instance Eq CHSHook where
 data CHSTrans = CHSTrans Bool			-- underscore to case?
 			 [(Ident, Ident)]	-- alias list
 
+-- marshalling descriptor for function hooks (EXPORTED)
+--
+-- * a marshaller consists of a function name and flag indicating whether it
+--   has to be executed in the IO monad
+--
+data CHSParm = CHSParm (Maybe (Ident, CHSArg))	-- "in" marshaller
+		       String			-- Haskell type
+		       Bool			-- C repr: two values?
+		       (Maybe (Ident, CHSArg))	-- "out" marshaller
+		       Position
+
+-- kinds of arguments in function hooks (EXPORTED)
+--
+data CHSArg = CHSValArg				-- plain value argument
+	    | CHSIOArg				-- reference argument
+	    | CHSVoidArg			-- no argument
+	    deriving (Eq)
+
 -- structure member access types (EXPORTED)
 --
 data CHSAccess = CHSSet				-- set structure field
@@ -212,6 +289,17 @@ instance Show CHSPtrType where
   show CHSForeignPtr	 = "ForeignPtr"
   show CHSStablePtr	 = "StablePtr"
 
+instance Read CHSPtrType where
+  readsPrec _ (                            'P':'t':'r':rest) = 
+    [(CHSPtr, rest)]
+  readsPrec _ ('F':'o':'r':'e':'i':'g':'n':'P':'t':'r':rest) = 
+    [(CHSForeignPtr, rest)]
+  readsPrec _ ('S':'t':'a':'b':'l':'e'    :'P':'t':'r':rest) = 
+    [(CHSStablePtr, rest)]
+  readsPrec p (c:cs)
+    | isSpace c						     = readsPrec p cs
+  readsPrec _ _						     = []
+
 
 -- load and dump a CHS file
 -- ------------------------
@@ -289,35 +377,69 @@ dumpCHS fname mod pureHaskell  =
       \-- ** Edit the orignal .chs file instead!\n\n"
       ++ showCHSModule mod pureHaskell
 
+-- to keep track of the current state of the line emission automaton
+--
+data LineState = Emit		-- emit LINE pragma if next frag is Haskell
+	       | Wait		-- emit LINE pragma after the next '\n'
+	       | NoLine		-- no pragma needed
+	       deriving (Eq)
+
 -- convert a CHS module into a string
 --
--- * if the second argument is `True', it is all Haskell code
+-- * if the second argument is `True', all fragments must contain Haskell code
 --
 showCHSModule                               :: CHSModule -> Bool -> String
 showCHSModule (CHSModule frags) pureHaskell  = 
-  (  foldr (.) id 
-   . map (showCHSFrag pureHaskell) 
-   $ frags) ""
-
-showCHSFrag :: Bool -> CHSFrag -> ShowS
-showCHSFrag _     (CHSVerb s   ) = showString s
-showCHSFrag False (CHSHook hook) =   showString "{#" 
-				   . showCHSHook hook
-				   . showString "#}"
-showCHSFrag True  (CHSHook _   ) = interr "showCHSFrag: Illegal hook!"
+  showFrags pureHaskell Emit frags []
+  where
+    -- the second argument indicates whether the next fragment (if it is
+    -- Haskell code) should be preceded by a LINE pragma; in particular
+    -- generated fragments and those following them need to be prefixed with a
+    -- LINE pragma
+    --
+    showFrags :: Bool -> LineState -> [CHSFrag] -> ShowS
+    showFrags _      _     []                           = id
+    showFrags pureHs state (CHSVerb s      pos : frags) = 
+      let
+	(fname, line, _) = pos
+	generated	 = isBuiltinPos pos
+	emitNow		 = state == Emit || 
+			   (state == Wait && not (null s) && head s == '\n')
+	nextState	 = if generated then Wait else NoLine
+      in
+	(if emitNow then
+	   showString ("{-# LINE " ++ show (line `max` 0) ++ " " ++ show fname 
+		       ++ " #-}\n")
+	 else id)
+      . showString s
+      . showFrags pureHs nextState frags
+    showFrags False  _     (CHSHook hook       : frags) =   
+        showString "{#" 
+      . showCHSHook hook
+      . showString "#}"
+      . showFrags False Wait frags
+    showFrags False  _     (CHSCPP  s    _     : frags) =   
+        showChar '#'
+      . showString s
+      . showChar '\n'
+      . showFrags False Emit frags
+    showFrags False  _     (CHSC    s    _     : frags) =
+        showString "\n#c\n"
+      . showString s
+      . showString "\n#endc\n"
+      . showFrags False Emit frags
+    showFrags False  _     (CHSCond _    _     : frags) =
+      interr "showCHSFrag: Cannot print `CHSCond'!"
+    showFrags True   _     _                            =
+      interr "showCHSFrag: Illegal hook, cpp directive, or inline C code!"
 
 showCHSHook :: CHSHook -> ShowS
 showCHSHook (CHSImport isQual ide _ _) =   
     showString "import "
   . (if isQual then showString "qualified " else id)
   . showCHSIdent ide
-showCHSHook (CHSContext oheader olib oprefix _) =   
+showCHSHook (CHSContext olib oprefix _) =   
     showString "context "
-  . (case oheader of
-       Nothing     -> showString ""
-       Just header ->   showString "header = " 
-		      . showString header 
-		      . showString " ")
   . (case olib of
        Nothing  -> showString ""
        Just lib -> showString "lib = " . showString lib . showString " ")
@@ -337,11 +459,23 @@ showCHSHook (CHSEnum ide oalias trans oprefix derive _) =
       "deriving (" 
       ++ concat (intersperse ", " (map identToLexeme derive))
       ++ ") "
-showCHSHook (CHSCall isFun isUns ide oalias _) =   
+showCHSHook (CHSCall isPure isUns ide oalias _) =   
     showString "call "
-  . (if isFun then showString "fun " else id)
+  . (if isPure then showString "pure " else id)
   . (if isUns then showString "unsafe " else id)
   . showIdAlias ide oalias
+showCHSHook (CHSFun isPure isUns ide oalias octxt parms parm _) =   
+    showString "fun "
+  . (if isPure then showString "pure " else id)
+  . (if isUns then showString "unsafe " else id)
+  . showIdAlias ide oalias
+  . (case octxt of
+       Nothing      -> showChar ' '
+       Just ctxtStr -> showString ctxtStr . showString " => ")
+  . showString "{"
+  . foldr (.) id (intersperse (showString ", ") (map showCHSParm parms))
+  . showString "} -> "
+  . showCHSParm parm
 showCHSHook (CHSField acc path _) =   
     (case acc of
        CHSGet -> showString "get "
@@ -359,6 +493,14 @@ showCHSHook (CHSPointer star ide oalias ptrType isNewtype oRefType _) =
        (True , _       ) -> showString " newtype" 
        (False, Just ide) -> showString " -> " . showCHSIdent ide
        (False, Nothing ) -> showString "")
+showCHSHook (CHSClass oclassIde classIde typeIde _) =   
+    showString "class "
+  . (case oclassIde of
+       Nothing       -> showString ""
+       Just classIde -> showCHSIdent classIde . showString " => ")
+  . showCHSIdent classIde
+  . showString " "
+  . showCHSIdent typeIde
 
 showPrefix                        :: Maybe String -> Bool -> ShowS
 showPrefix Nothing       _         = showString ""
@@ -373,23 +515,36 @@ showIdAlias            :: Ident -> Maybe Ident -> ShowS
 showIdAlias ide oalias  =
     showCHSIdent ide
   . (case oalias of
-       Nothing  -> showString ""
+       Nothing  -> id
        Just ide -> showString " as " . showCHSIdent ide)
 
-showCHSTrans :: CHSTrans -> ShowS
-showCHSTrans (CHSTrans _2Case assocs) =   
+showCHSParm                                                :: CHSParm -> ShowS
+showCHSParm (CHSParm oimMarsh hsTyStr twoCVals oomMarsh _)  =
+    showOMarsh oimMarsh
+  . showChar ' '
+  . showHsVerb hsTyStr
+  . (if twoCVals then showChar '&' else id)
+  . showChar ' '
+  . showOMarsh oomMarsh
+  where
+    showOMarsh Nothing               = id
+    showOMarsh (Just (ide, argKind)) =   showCHSIdent ide
+				       . (case argKind of
+					   CHSValArg  -> id
+					   CHSIOArg   -> showString "*"
+					   CHSVoidArg -> showString "-")
+    --
+    showHsVerb str = showChar '`' . showString str . showChar '\''
+
+showCHSTrans                          :: CHSTrans -> ShowS
+showCHSTrans (CHSTrans _2Case assocs)  =   
     showString "{"
   . (if _2Case then showString ("underscoreToCase" ++ maybeComma) else id)
-  . showAssocs assocs _2Case
+  . foldr (.) id (intersperse (showString ", ") (map showAssoc assocs))
   . showString "}"
   where
     maybeComma = if null assocs then "" else ", "
     --
-    showAssocs []             _      = id
-    showAssocs (assoc:assocs) notFst =
-	(if notFst then showString ", " else id)
-      . showAssoc assoc
-      . showAssocs assocs True
     showAssoc (ide1, ide2) =
 	showCHSIdent ide1
       . showString " as "
@@ -541,29 +696,39 @@ parseCHSModule pos cs  = do
 -- * in case of an error, all tokens that are neither Haskell nor control
 --   tokens are skipped; afterwards parsing continues
 --
+-- * when encountering inline-C code we scan forward over all inline-C and
+--   control tokens to avoid turning the control tokens within a sequence of
+--   inline-C into Haskell fragments
+--
 parseFrags      :: [CHSToken] -> CST s [CHSFrag]
 parseFrags toks  = do
 		     parseFrags0 toks
 		     `ifError` contFrags toks
   where
     parseFrags0 :: [CHSToken] -> CST s [CHSFrag]
-    parseFrags0 []                       = return []
-    parseFrags0 (CHSTokHaskell _ s:toks) = do
-					     frags <- parseFrags toks
-					     return (CHSVerb s : frags)
-    parseFrags0 (CHSTokCtrl    _ c:toks) = do
-					     frags <- parseFrags toks
-					     return (CHSVerb [c] : frags)
-    parseFrags0 (CHSTokImport  pos:toks) = parseImport  pos        toks
-    parseFrags0 (CHSTokContext pos:toks) = parseContext pos        toks
-    parseFrags0 (CHSTokType    pos:toks) = parseType    pos        toks
-    parseFrags0 (CHSTokSizeof  pos:toks) = parseSizeof  pos        toks
-    parseFrags0 (CHSTokEnum    pos:toks) = parseEnum    pos        toks
-    parseFrags0 (CHSTokCall    pos:toks) = parseCall    pos        toks
-    parseFrags0 (CHSTokGet     pos:toks) = parseField   pos CHSGet toks
-    parseFrags0 (CHSTokSet     pos:toks) = parseField   pos CHSSet toks
-    parseFrags0 (CHSTokPointer pos:toks) = parsePointer pos        toks
-    parseFrags0 toks			 = syntaxError toks
+    parseFrags0 []                         = return []
+    parseFrags0 (CHSTokHaskell pos s:toks) = do
+					       frags <- parseFrags toks
+					       return $ CHSVerb s pos : frags
+    parseFrags0 (CHSTokCtrl    pos c:toks) = do
+					       frags <- parseFrags toks
+					       return $ CHSVerb [c] pos : frags
+    parseFrags0 (CHSTokCPP     pos s:toks) = do
+					       frags <- parseFrags toks
+					       return $ CHSCPP s pos : frags
+    parseFrags0 (CHSTokC       pos s:toks) = parseC       pos s      toks 
+    parseFrags0 (CHSTokImport  pos  :toks) = parseImport  pos        toks
+    parseFrags0 (CHSTokContext pos  :toks) = parseContext pos        toks
+    parseFrags0 (CHSTokType    pos  :toks) = parseType    pos        toks
+    parseFrags0 (CHSTokSizeof  pos  :toks) = parseSizeof  pos        toks
+    parseFrags0 (CHSTokEnum    pos  :toks) = parseEnum    pos        toks
+    parseFrags0 (CHSTokCall    pos  :toks) = parseCall    pos        toks
+    parseFrags0 (CHSTokFun     pos  :toks) = parseFun     pos        toks
+    parseFrags0 (CHSTokGet     pos  :toks) = parseField   pos CHSGet toks
+    parseFrags0 (CHSTokSet     pos  :toks) = parseField   pos CHSSet toks
+    parseFrags0 (CHSTokClass   pos  :toks) = parseClass   pos        toks
+    parseFrags0 (CHSTokPointer pos  :toks) = parsePointer pos        toks
+    parseFrags0 toks			   = syntaxError toks
     --
     -- skip to next Haskell or control token
     --
@@ -572,6 +737,20 @@ parseFrags toks  = do
     contFrags toks@(CHSTokCtrl    _ _:_   ) = parseFrags toks
     contFrags      (_                :toks) = contFrags  toks
 
+parseC :: Position -> String -> [CHSToken] -> CST s [CHSFrag]
+parseC pos s toks = 
+  do
+    frags <- collectCtrlAndC toks
+    return $ CHSC s pos : frags
+  where
+    collectCtrlAndC (CHSTokCtrl pos c:toks) = do
+					        frags <- collectCtrlAndC toks
+						return $ CHSC [c] pos : frags
+    collectCtrlAndC (CHSTokC    pos s:toks) = do
+					        frags <- collectCtrlAndC toks
+						return $ CHSC s   pos : frags
+    collectCtrlAndC toks		    = parseFrags toks
+
 parseImport :: Position -> [CHSToken] -> CST s [CHSFrag]
 parseImport pos toks = do
   (qual, modid, toks') <- 
@@ -586,12 +765,11 @@ parseImport pos toks = do
 
 parseContext          :: Position -> [CHSToken] -> CST s [CHSFrag]
 parseContext pos toks  = do
-		           (oheader , toks' )  <- parseOptHeader       toks
-		           (olib    , toks'' ) <- parseOptLib          toks'
+		           (olib    , toks'' ) <- parseOptLib          toks
 		           (opref   , toks''') <- parseOptPrefix False toks''
 		           toks''''	       <- parseEndHook	       toks'''
 		           frags               <- parseFrags           toks''''
-			   let frag = CHSContext oheader olib opref pos
+			   let frag = CHSContext olib opref pos
 		           return $ CHSHook frag : frags
 
 parseType :: Position -> [CHSToken] -> CST s [CHSFrag]
@@ -613,7 +791,7 @@ parseSizeof _ toks = syntaxError toks
 parseEnum :: Position -> [CHSToken] -> CST s [CHSFrag]
 parseEnum pos (CHSTokIdent _ ide:toks) =
   do
-    (oalias, toks' )   <- parseOptAs          toks
+    (oalias, toks' )   <- parseOptAs ide True toks
     (trans , toks'')   <- parseTrans          toks'
     (oprefix, toks''') <- parseOptPrefix True toks''
     (derive, toks'''') <- parseDerive	      toks'''
@@ -629,23 +807,91 @@ parseEnum _ toks = syntaxError toks
 parseCall          :: Position -> [CHSToken] -> CST s [CHSFrag]
 parseCall pos toks  = 
   do
-    (isFun   , toks'   ) <- parseIsFun    toks
-    (isUnsafe, toks''  ) <- parseIsUnsafe toks'
-    (ide     , toks''' ) <- parseIdent    toks''
-    (oalias  , toks'''') <- parseOptAs    toks'''
-    toks'''''		 <- parseEndHook  toks''''
-    frags                <- parseFrags    toks'''''
-    return $ CHSHook (CHSCall isFun isUnsafe ide (norm ide oalias) pos) : frags
+    (isPure  , toks'   ) <- parseIsPure          toks
+    (isUnsafe, toks''  ) <- parseIsUnsafe        toks'
+    (ide     , toks''' ) <- parseIdent           toks''
+    (oalias  , toks'''') <- parseOptAs ide False toks'''
+    toks'''''		 <- parseEndHook	 toks''''
+    frags                <- parseFrags		 toks'''''
+    return $ 
+      CHSHook (CHSCall isPure isUnsafe ide (norm ide oalias) pos) : frags
+
+parseFun          :: Position -> [CHSToken] -> CST s [CHSFrag]
+parseFun pos toks  = 
+  do
+    (isPure  , toks' ) <- parseIsPure          toks
+    (isUnsafe, toks'2) <- parseIsUnsafe        toks'
+    (ide     , toks'3) <- parseIdent           toks'2
+    (oalias  , toks'4) <- parseOptAs ide False toks'3
+    (octxt   , toks'5) <- parseOptContext      toks'4
+    (parms   , toks'6) <- parseParms	       toks'5
+    (parm    , toks'7) <- parseParm	       toks'6
+    toks'8	       <- parseEndHook	       toks'7
+    frags              <- parseFrags	       toks'8
+    return $ 
+      CHSHook 
+        (CHSFun isPure isUnsafe ide (norm ide oalias) octxt parms parm pos) :
+      frags
   where
-    parseIsFun (CHSTokFun _:toks) = return (True, toks)
-    parseIsFun toks		  = return (False, toks)
+    parseOptContext (CHSTokHSVerb _ ctxt:CHSTokDArrow _:toks) =
+      return (Just ctxt, toks)
+    parseOptContext toks				      =
+      return (Nothing  , toks)
+    --
+    parseParms (CHSTokLBrace _:CHSTokRBrace _:CHSTokArrow _:toks) = 
+      return ([], toks)
+    parseParms (CHSTokLBrace _                             :toks) = 
+      parseParms' (CHSTokComma nopos:toks)
+    parseParms						    toks  = 
+      syntaxError toks
+    --
+    parseParms' (CHSTokRBrace _:CHSTokArrow _:toks) = return ([], toks)
+    parseParms' (CHSTokComma _               :toks) = do
+      (parm , toks' ) <- parseParm   toks
+      (parms, toks'') <- parseParms' toks'
+      return (parm:parms, toks'')
+    parseParms' (CHSTokRBrace _              :toks) = syntaxError toks
+      -- gives better error messages
+    parseParms'				      toks  = syntaxError toks
 
-    parseIsUnsafe (CHSTokUnsafe _:toks) = return (True, toks)
-    parseIsUnsafe toks		        = return (False, toks)
+parseIsPure :: [CHSToken] -> CST s (Bool, [CHSToken])
+parseIsPure (CHSTokPure _:toks) = return (True , toks)
+parseIsPure (CHSTokFun  _:toks) = return (True , toks)  -- backwards compat.
+parseIsPure toks		= return (False, toks)
+-- FIXME: eventually, remove `fun'; it's currently deprecated
 
-    norm ide Nothing                   = Nothing
-    norm ide (Just ide') | ide == ide' = Nothing
-			 | otherwise   = Just ide'
+parseIsUnsafe :: [CHSToken] -> CST s (Bool, [CHSToken])
+parseIsUnsafe (CHSTokUnsafe _:toks) = return (True , toks)
+parseIsUnsafe toks		    = return (False, toks)
+
+norm :: Ident -> Maybe Ident -> Maybe Ident
+norm ide Nothing                   = Nothing
+norm ide (Just ide') | ide == ide' = Nothing
+		     | otherwise   = Just ide'
+
+parseParm :: [CHSToken] -> CST s (CHSParm, [CHSToken])
+parseParm toks =
+  do
+    (oimMarsh, toks' ) <- parseOptMarsh toks
+    (hsTyStr, twoCVals, pos, toks'2) <- 
+      case toks' of
+	(CHSTokHSVerb pos hsTyStr:CHSTokAmp _:toks'2) -> 
+	  return (hsTyStr, True , pos, toks'2)
+	(CHSTokHSVerb pos hsTyStr	     :toks'2) -> 
+	  return (hsTyStr, False, pos, toks'2)
+	toks				              -> syntaxError toks
+    (oomMarsh, toks'3) <- parseOptMarsh toks'2
+    return (CHSParm oimMarsh hsTyStr twoCVals oomMarsh pos, toks'3)
+  where
+    parseOptMarsh :: [CHSToken] -> CST s (Maybe (Ident, CHSArg), [CHSToken])
+    parseOptMarsh (CHSTokIdent _ ide:CHSTokStar _ :toks) = 
+      return (Just (ide, CHSIOArg) , toks)
+    parseOptMarsh (CHSTokIdent _ ide:CHSTokMinus _:toks) = 
+      return (Just (ide, CHSVoidArg), toks)
+    parseOptMarsh (CHSTokIdent _ ide              :toks) = 
+      return (Just (ide, CHSValArg) , toks)
+    parseOptMarsh toks					 =
+      return (Nothing, toks)
 
 parseField :: Position -> CHSAccess -> [CHSToken] -> CST s [CHSFrag]
 parseField pos access toks =
@@ -662,8 +908,8 @@ parsePointer pos toks =
 	CHSTokStar _:CHSTokIdent _ ide:toks' -> return (True , ide, toks')
 	CHSTokIdent _ ide             :toks' -> return (False, ide, toks')
 	_				     -> syntaxError toks
-    (oalias , toks'2)             <- parseOptAs   toks'
-    (ptrType, toks'3)             <- parsePtrType toks'2
+    (oalias , toks'2)             <- parseOptAs ide True toks'
+    (ptrType, toks'3)             <- parsePtrType        toks'2
     let 
      (isNewtype, oRefType, toks'4) =
       case toks'3 of
@@ -686,13 +932,24 @@ parsePointer pos toks =
     norm ide (Just ide') | ide == ide' = Nothing
 			 | otherwise   = Just ide'
 
-parseOptHeader :: [CHSToken] -> CST s (Maybe String, [CHSToken])
-parseOptHeader (CHSTokHeader _    :
-	        CHSTokEqual  _    :
-	        CHSTokString _ str:
-	        toks)	             = return (Just str, toks)
-parseOptHeader (CHSTokHeader _:toks) = syntaxError toks
-parseOptHeader toks		     = return (Nothing, toks)
+parseClass :: Position -> [CHSToken] -> CST s [CHSFrag]
+parseClass pos (CHSTokIdent  _ sclassIde:
+	        CHSTokDArrow _          :
+		CHSTokIdent  _ classIde :
+		CHSTokIdent  _ typeIde  :
+		toks)                     =
+  do
+    toks' <- parseEndHook toks
+    frags <- parseFrags toks'
+    return $ CHSHook (CHSClass (Just sclassIde) classIde typeIde pos) : frags
+parseClass pos (CHSTokIdent _ classIde :
+		CHSTokIdent _ typeIde  :
+		toks)                     =
+  do
+    toks' <- parseEndHook toks
+    frags <- parseFrags toks'
+    return $ CHSHook (CHSClass Nothing classIde typeIde pos) : frags
+parseClass _ toks = syntaxError toks
 
 parseOptLib :: [CHSToken] -> CST s (Maybe String, [CHSToken])
 parseOptLib (CHSTokLib    _    :
@@ -716,11 +973,37 @@ parseOptPrefix _     (CHSTokWith   _:toks) = syntaxError toks
 parseOptPrefix _     (CHSTokPrefix _:toks) = syntaxError toks
 parseOptPrefix _     toks		   = return (Nothing, toks)
 
-parseOptAs :: [CHSToken] -> CST s (Maybe Ident, [CHSToken])
-parseOptAs (CHSTokAs _:CHSTokIdent _ ide:toks) = return (Just ide, toks)
-parseOptAs (CHSTokAs _:toks		     ) = syntaxError toks
-parseOptAs  toks			       = return (Nothing, toks)
+-- first argument is the identifier that is to be used when `^' is given and
+-- the second indicates whether the first characters has to be upper case
+--
+parseOptAs :: Ident -> Bool -> [CHSToken] -> CST s (Maybe Ident, [CHSToken])
+parseOptAs _   _     (CHSTokAs _:CHSTokIdent _ ide:toks) = 
+  return (Just ide, toks)
+parseOptAs ide upper (CHSTokAs _:CHSTokHat pos    :toks) = 
+  return (Just $ underscoreToCase ide upper pos, toks)
+parseOptAs _   _     (CHSTokAs _                  :toks) = syntaxError toks
+parseOptAs _   _				   toks	 = 
+  return (Nothing, toks)
 
+-- convert C style identifier to Haskell style identifier
+--
+underscoreToCase               :: Ident -> Bool -> Position -> Ident
+underscoreToCase ide upper pos  = 
+  let lexeme = identToLexeme ide
+      ps     = filter (not . null) . parts $ lexeme
+  in
+  onlyPosIdent pos . adjustHead . concat . map adjustCase $ ps
+  where
+    parts s = let (l, s') = break (== '_') s
+	      in  
+	      l : case s' of
+		    []      -> []
+		    (_:s'') -> parts s''
+    --    
+    adjustCase (c:cs) = toUpper c : map toLower cs
+    --
+    adjustHead ""     = ""
+    adjustHead (c:cs) = if upper then toUpper c : cs else toLower c:cs
 
 -- this is disambiguated and left factored
 --
