@@ -1,11 +1,11 @@
 --  C->Haskell Compiler: Lexer for CHS Files
 --
---  Author : Manuel M. T. Chakravarty
+--  Author : Manuel M T Chakravarty
 --  Created: 13 August 99
 --
---  Version $Revision: 1.12 $ from $Date: 2001/06/20 09:25:13 $
+--  Version $Revision: 1.19 $ from $Date: 2003/04/16 11:13:10 $
 --
---  Copyright (c) [1999..2001] Manuel M. T. Chakravarty
+--  Copyright (c) [1999..2003] Manuel M T Chakravarty
 --
 --  This file is free software; you can redistribute it and/or modify
 --  it under the terms of the GNU General Public License as published by
@@ -33,7 +33,8 @@
 --    (realised as two lexer coupled by meta actions) depending on whether
 --    Haskell code or a binding hook is currently read.  The lexer reading
 --    Haskell code is called `base lexer'; the other one, `binding-hook
---    lexer'.
+--    lexer'.  In addition, there is a inline-c lexer, which, as the
+--    binding-hook lexer, can be triggered from the base lexer.
 --
 --  * Base lexer:
 --
@@ -42,6 +43,8 @@
 --		 | comment
 --		 | nested
 --		 | hstring
+--		 | '{#'
+--		 | cpp
 --      special -> `(' | `{' | `-' | `"'
 --      ctrl    -> `\n' | `\f' | `\r' | `\t' | `\v'
 --
@@ -51,7 +54,7 @@
 --    Within the base lexer control codes appear as separate tokens in the
 --    token list.
 --
---    NOTE: It is important that `(' is an extra lexeme and not added as an
+--    NOTE: It is important that `{' is an extra lexeme and not added as an
 --	    optional component at the end of the first alternative for
 --	    `haskell'.  Otherwise, the principle of the longest match will
 --	    divide `foo {#' into the tokens `foo {' and `#' instead of `foo '
@@ -71,17 +74,30 @@
 --      inhstr  -> ` '..`\127' \\ `"'
 --		 | `\"'
 --
+--    Pre-precessor directives as well as the switch to inline-C code are
+--    formed as follows:
+--
+--      cpp	-> `\n#' (inline | `\t')* `\n'
+--		 | `\n#c' (' ' | '\t')* `\n'
+--
+--    We allow whitespace between the `#' and the actual directive, but in `#c'
+--    and `#endc' the directive must immediately follow the `#'.  This might
+--    be regarded as a not entirely orthogonal design, but simplifies matters
+--    especially for `#endc'.
+--
 --  * On encountering the lexeme `{#', a meta action in the base lexer
 --    transfers control to the following binding-hook lexer:
 --
---      ident       -> letter (letter | digit | ')*
---      reservedid  -> `as' | `call' | `context' | `deriving' 
---		     | `enum' | `foreign' | `fun' | `get' | `header' | `lib' 
---		     | `newtype' | `pointer' | `prefix' | `set' | `sizeof'
---		     | `stable' | `type' | `underscoreToCase' | `unsafe' 
---		     | `with'
---      reservedsym -> `{#' | `#}' | `{' | `}' | `,' | `.' | `->' | `=' | `*'
+--      ident       -> letter (letter | digit | `'')*
+--      reservedid  -> `as' | `call' | `class' | `context' | `deriving' 
+--		     | `enum' | `foreign' | `fun' | `get' | `lib' 
+--		     | `newtype' | `pointer' | `prefix' | `pure' | `set'
+--		     | `sizeof' | `stable' | `type' | `underscoreToCase' 
+--		     | `unsafe' | `with'
+--      reservedsym -> `{#' | `#}' | `{' | `}' | `,' | `.' | `->' | `=' 
+--		     | `=>' | '-' | `*' | `&' | `^'
 --      string      -> `"' instr* `"'
+--      verbhs      -> `\`' instr* `\''
 --      instr       -> ` '..`\127' \\ `"'
 --      comment	    -> `--' (any \\ `\n')* `\n'
 --
@@ -104,12 +120,37 @@
 --    starting with a lowercase and those starting with an uppercase letter).
 --    However, we use it also to scan C identifiers; although, strictly
 --    speaking, it is too general for them.  In the case of C identifiers,
---    this should not have any impact on the range of descirptions accepted by
+--    this should not have any impact on the range of descriptions accepted by
 --    the tool, as illegal identifier will never occur in a C header file that
 --    is accepted by the C lexer.  In the case of Haskell identifiers, a
 --    confusion between variable and constructor identifiers will be noted by
 --    the Haskell compiler translating the code generated by c2hs.
 --
+--  * Any line starting with the character `#' is regarded to be a C
+--    preprocessor directive.  With the exception of `#c' and `#endc', which
+--    delimit a set of lines containing inline C code.  Hence, in the base
+--    lexer, the lexeme `#c' triggers a meta action transferring control to the
+--    following inline-C lexer:
+--
+--      c  -> inline* \\ `\n#endc'
+--
+--    We do neither treat C strings nor C comments specially.  Hence, if the
+--    string "\n#endc" occurs in a comment, we will mistakenly regard it as
+--    the end of the inline C code.  Note that the problem cannot happen with
+--    strings, as C does not permit strings that extend over multiple lines.
+--    At the moment, it just seems not to be worth the effort required to
+--    treat this situation more accurately.
+--
+--    The inline-C lexer also doesn't handle pre-processor directives
+--    specially.  Hence, structural pre-processor directives (namely,
+--    conditionals) may occur within inline-C code only properly nested.
+--
+--  Shortcomings
+--  ~~~~~~~~~~~~
+--  Some lexemes that include single and double quote characters are not lexed
+--  correctly.  See the implementation comment at `haskell' for details.
+--
+--
 --- TODO ----------------------------------------------------------------------
 --
 --  * In `haskell', the case of a single `"' (without a matching second one)
@@ -117,7 +158,7 @@
 --    necessary, but for some strange reason, the lexer otherwise hangs when a 
 --    single `"' appears in the input.
 --
---  * Comments in the "gap" of a string a not yet supported.
+--  * Comments in the "gap" of a string are not yet supported.
 --
 
 module CHSLexer (CHSToken(..), lexCHS) 
@@ -144,10 +185,14 @@ import C2HSState (CST, raise, raiseError, nop, getNameSupply)
 -- possible tokens (EXPORTED)
 --
 data CHSToken = CHSTokArrow   Position		-- `->'
+	      | CHSTokDArrow  Position		-- `=>'
 	      | CHSTokDot     Position		-- `.'
 	      | CHSTokComma   Position		-- `,'
 	      | CHSTokEqual   Position		-- `='
+	      | CHSTokMinus   Position		-- `-'
 	      | CHSTokStar    Position		-- `*'
+	      | CHSTokAmp     Position		-- `&'
+	      | CHSTokHat     Position		-- `^'
 	      | CHSTokLBrace  Position		-- `{'
 	      | CHSTokRBrace  Position		-- `}'
 	      | CHSTokLParen  Position		-- `('
@@ -155,18 +200,19 @@ data CHSToken = CHSTokArrow   Position		-- `->'
 	      | CHSTokEndHook Position		-- `#}'
 	      | CHSTokAs      Position		-- `as'
 	      | CHSTokCall    Position		-- `call'
+	      | CHSTokClass   Position		-- `class'
 	      | CHSTokContext Position		-- `context'
 	      | CHSTokDerive  Position		-- `deriving'
 	      | CHSTokEnum    Position		-- `enum'
 	      | CHSTokForeign Position          -- `foreign'
 	      | CHSTokFun     Position		-- `fun'
 	      | CHSTokGet     Position		-- `get'
-	      | CHSTokHeader  Position		-- `header'
 	      | CHSTokImport  Position		-- `import'
 	      | CHSTokLib     Position		-- `lib'
 	      | CHSTokNewtype Position		-- `newtype'
 	      | CHSTokPointer Position		-- `pointer'
 	      | CHSTokPrefix  Position		-- `prefix'
+	      | CHSTokPure    Position		-- `pure'
 	      | CHSTokQualif  Position		-- `qualified'
 	      | CHSTokSet     Position		-- `set'
 	      | CHSTokSizeof  Position		-- `sizeof'
@@ -176,16 +222,23 @@ data CHSToken = CHSTokArrow   Position		-- `->'
 	      | CHSTokUnsafe  Position		-- `unsafe'
 	      | CHSTokWith    Position		-- `with'
 	      | CHSTokString  Position String	-- string 
+	      | CHSTokHSVerb  Position String	-- verbatim Haskell (`...')
 	      | CHSTokIdent   Position Ident	-- identifier
 	      | CHSTokHaskell Position String	-- verbatim Haskell code
+	      | CHSTokCPP     Position String	-- pre-processor directive
+	      | CHSTokC	      Position String	-- verbatim C code
 	      | CHSTokCtrl    Position Char	-- control code
 
 instance Pos CHSToken where
   posOf (CHSTokArrow   pos  ) = pos
+  posOf (CHSTokDArrow  pos  ) = pos
   posOf (CHSTokDot     pos  ) = pos
   posOf (CHSTokComma   pos  ) = pos
   posOf (CHSTokEqual   pos  ) = pos
+  posOf (CHSTokMinus   pos  ) = pos
   posOf (CHSTokStar    pos  ) = pos
+  posOf (CHSTokAmp     pos  ) = pos
+  posOf (CHSTokHat     pos  ) = pos
   posOf (CHSTokLBrace  pos  ) = pos
   posOf (CHSTokRBrace  pos  ) = pos
   posOf (CHSTokLParen  pos  ) = pos
@@ -193,18 +246,19 @@ instance Pos CHSToken where
   posOf (CHSTokEndHook pos  ) = pos
   posOf (CHSTokAs      pos  ) = pos
   posOf (CHSTokCall    pos  ) = pos
+  posOf (CHSTokClass   pos  ) = pos
   posOf (CHSTokContext pos  ) = pos
   posOf (CHSTokDerive  pos  ) = pos
   posOf (CHSTokEnum    pos  ) = pos
   posOf (CHSTokForeign pos  ) = pos
   posOf (CHSTokFun     pos  ) = pos
   posOf (CHSTokGet     pos  ) = pos
-  posOf (CHSTokHeader  pos  ) = pos
   posOf (CHSTokImport  pos  ) = pos
   posOf (CHSTokLib     pos  ) = pos
   posOf (CHSTokNewtype pos  ) = pos
   posOf (CHSTokPointer pos  ) = pos
   posOf (CHSTokPrefix  pos  ) = pos
+  posOf (CHSTokPure    pos  ) = pos
   posOf (CHSTokQualif  pos  ) = pos
   posOf (CHSTokSet     pos  ) = pos
   posOf (CHSTokSizeof  pos  ) = pos
@@ -214,16 +268,23 @@ instance Pos CHSToken where
   posOf (CHSTokUnsafe  pos  ) = pos
   posOf (CHSTokWith    pos  ) = pos
   posOf (CHSTokString  pos _) = pos
+  posOf (CHSTokHSVerb  pos _) = pos
   posOf (CHSTokIdent   pos _) = pos
   posOf (CHSTokHaskell pos _) = pos
+  posOf (CHSTokCPP     pos _) = pos
+  posOf (CHSTokC       pos _) = pos
   posOf (CHSTokCtrl    pos _) = pos
 
 instance Eq CHSToken where
   (CHSTokArrow    _  ) == (CHSTokArrow    _  ) = True
+  (CHSTokDArrow   _  ) == (CHSTokDArrow   _  ) = True
   (CHSTokDot      _  ) == (CHSTokDot      _  ) = True
   (CHSTokComma    _  ) == (CHSTokComma    _  ) = True
   (CHSTokEqual    _  ) == (CHSTokEqual    _  ) = True
+  (CHSTokMinus    _  ) == (CHSTokMinus    _  ) = True
   (CHSTokStar     _  ) == (CHSTokStar     _  ) = True
+  (CHSTokAmp      _  ) == (CHSTokAmp      _  ) = True
+  (CHSTokHat      _  ) == (CHSTokHat      _  ) = True
   (CHSTokLBrace   _  ) == (CHSTokLBrace   _  ) = True
   (CHSTokRBrace   _  ) == (CHSTokRBrace   _  ) = True
   (CHSTokLParen   _  ) == (CHSTokLParen   _  ) = True
@@ -231,18 +292,19 @@ instance Eq CHSToken where
   (CHSTokEndHook  _  ) == (CHSTokEndHook  _  ) = True
   (CHSTokAs       _  ) == (CHSTokAs       _  ) = True
   (CHSTokCall     _  ) == (CHSTokCall     _  ) = True
+  (CHSTokClass    _  ) == (CHSTokClass    _  ) = True
   (CHSTokContext  _  ) == (CHSTokContext  _  ) = True
   (CHSTokDerive   _  ) == (CHSTokDerive   _  ) = True
   (CHSTokEnum     _  ) == (CHSTokEnum     _  ) = True
   (CHSTokForeign  _  ) == (CHSTokForeign  _  ) = True
   (CHSTokFun      _  ) == (CHSTokFun      _  ) = True
   (CHSTokGet      _  ) == (CHSTokGet      _  ) = True
-  (CHSTokHeader   _  ) == (CHSTokHeader   _  ) = True
   (CHSTokImport   _  ) == (CHSTokImport   _  ) = True
   (CHSTokLib      _  ) == (CHSTokLib      _  ) = True
   (CHSTokNewtype  _  ) == (CHSTokNewtype  _  ) = True
   (CHSTokPointer  _  ) == (CHSTokPointer  _  ) = True
   (CHSTokPrefix   _  ) == (CHSTokPrefix   _  ) = True
+  (CHSTokPure     _  ) == (CHSTokPure     _  ) = True
   (CHSTokQualif   _  ) == (CHSTokQualif   _  ) = True
   (CHSTokSet      _  ) == (CHSTokSet      _  ) = True
   (CHSTokSizeof   _  ) == (CHSTokSizeof   _  ) = True
@@ -252,17 +314,24 @@ instance Eq CHSToken where
   (CHSTokUnsafe   _  ) == (CHSTokUnsafe   _  ) = True
   (CHSTokWith     _  ) == (CHSTokWith     _  ) = True
   (CHSTokString   _ _) == (CHSTokString   _ _) = True
+  (CHSTokHSVerb   _ _) == (CHSTokHSVerb   _ _) = True
   (CHSTokIdent    _ _) == (CHSTokIdent    _ _) = True
   (CHSTokHaskell  _ _) == (CHSTokHaskell  _ _) = True
+  (CHSTokCPP	  _ _) == (CHSTokCPP	  _ _) = True
+  (CHSTokC	  _ _) == (CHSTokC	  _ _) = True
   (CHSTokCtrl	  _ _) == (CHSTokCtrl	  _ _) = True
   _		       == _		       = False
 
 instance Show CHSToken where
   showsPrec _ (CHSTokArrow   _  ) = showString "->"
+  showsPrec _ (CHSTokDArrow  _  ) = showString "=>"
   showsPrec _ (CHSTokDot     _  ) = showString "."
   showsPrec _ (CHSTokComma   _  ) = showString ","
   showsPrec _ (CHSTokEqual   _  ) = showString "="
+  showsPrec _ (CHSTokMinus   _  ) = showString "-"
   showsPrec _ (CHSTokStar    _  ) = showString "*"
+  showsPrec _ (CHSTokAmp     _  ) = showString "&"
+  showsPrec _ (CHSTokHat     _  ) = showString "^"
   showsPrec _ (CHSTokLBrace  _  ) = showString "{"
   showsPrec _ (CHSTokRBrace  _  ) = showString "}"
   showsPrec _ (CHSTokLParen  _  ) = showString "("
@@ -270,18 +339,19 @@ instance Show CHSToken where
   showsPrec _ (CHSTokEndHook _  ) = showString "#}"
   showsPrec _ (CHSTokAs      _  ) = showString "as"
   showsPrec _ (CHSTokCall    _  ) = showString "call"
+  showsPrec _ (CHSTokClass   _  ) = showString "class"
   showsPrec _ (CHSTokContext _  ) = showString "context"
   showsPrec _ (CHSTokDerive  _  ) = showString "deriving"
   showsPrec _ (CHSTokEnum    _  ) = showString "enum"
   showsPrec _ (CHSTokForeign _  ) = showString "foreign"
   showsPrec _ (CHSTokFun     _  ) = showString "fun"
   showsPrec _ (CHSTokGet     _  ) = showString "get"
-  showsPrec _ (CHSTokHeader  _  ) = showString "header"
   showsPrec _ (CHSTokImport  _  ) = showString "import"
   showsPrec _ (CHSTokLib     _  ) = showString "lib"
   showsPrec _ (CHSTokNewtype _  ) = showString "newtype"
   showsPrec _ (CHSTokPointer _  ) = showString "pointer"
   showsPrec _ (CHSTokPrefix  _  ) = showString "prefix"
+  showsPrec _ (CHSTokPure    _  ) = showString "pure"
   showsPrec _ (CHSTokQualif  _  ) = showString "qualified"
   showsPrec _ (CHSTokSet     _  ) = showString "set"
   showsPrec _ (CHSTokSizeof  _  ) = showString "sizeof"
@@ -291,8 +361,11 @@ instance Show CHSToken where
   showsPrec _ (CHSTokUnsafe  _  ) = showString "unsafe"
   showsPrec _ (CHSTokWith    _  ) = showString "with"
   showsPrec _ (CHSTokString  _ s) = showString ("\"" ++ s ++ "\"")
+  showsPrec _ (CHSTokHSVerb  _ s) = showString ("`" ++ s ++ "'")
   showsPrec _ (CHSTokIdent   _ i) = (showString . identToLexeme) i
   showsPrec _ (CHSTokHaskell _ s) = showString s
+  showsPrec _ (CHSTokCPP     _ s) = showString s
+  showsPrec _ (CHSTokC	     _ s) = showString s
   showsPrec _ (CHSTokCtrl    _ c) = showChar c
 
 
@@ -361,19 +434,30 @@ chslexer  =      haskell	-- Haskell code
 	    >||< nested		-- nested comments
 	    >||< ctrl		-- control code (that has to be preserved)
 	    >||< hook		-- start of a binding hook
+	    >||< cpp		-- a pre-processor directive (or `#c')
 
 -- stream of Haskell code (terminated by a control character or binding hook)
 --
 haskell :: CHSLexer
+--
+-- NB: We need to make sure that '"' is not regarded as the beginning of a
+--     string; however, we cannot really lex character literals properly
+--     without lexing identifiers (as the latter may containing single quotes
+--     as part of their lexeme).  Thus, we special case '"'.  This is still a
+--     kludge, as a program fragment, such as
+--
+--	 foo'"'strange string"
+--
+--     will not be handled correctly.
+--
 haskell  = (    anyButSpecial`star` epsilon
 	    >|< specialButQuotes
-	    >|< char '"' +> inhstr`star` char '"'
+	    >|< char '"'  +> inhstr`star` char '"'
+	    >|< string "'\"'"				-- special case of "
+	    >|< string "--" +> anyButNL`star` epsilon   -- comment
 	   )
 	   `lexaction` copyVerbatim
-	   >||< string "--" +> anyButNL`star` char '\n'	          -- comment
-                `lexmeta` (\cs pos s -> (Just $ Right (CHSTokHaskell pos cs),
-					 retPos pos, s, Nothing))
-	   >||< char '"'		             -- this is a bad kludge
+	   >||< char '"'		                -- this is a bad kludge
 		`lexactionErr` 
 		  \_ pos -> (Left $ makeError ErrorErr pos
 					      ["Lexical error!", 
@@ -382,8 +466,8 @@ haskell  = (    anyButSpecial`star` epsilon
 	     anyButSpecial    = alt (inlineSet \\ specialSet)
 	     specialButQuotes = alt (specialSet \\ ['"'])
 	     anyButNL	      = alt (anySet \\ ['\n'])
-	     inhstr	      = instr >|< string "\\\"" >|< gap
-	     gap	      = char '\\' +> alt (' ':ctrlSet)`star` char '\\'
+	     inhstr	      = instr >|< char '\\' >|< string "\\\"" >|< gap
+	     gap	      = char '\\' +> alt (' ':ctrlSet)`plus` char '\\'
 
 -- action copying the input verbatim to `CHSTokHaskell' tokens
 --
@@ -474,12 +558,31 @@ hook :: CHSLexer
 hook  = string "{#"
 	`lexmeta` \_ pos s -> (Nothing, incPos pos 2, s, Just bhLexer)
 
+-- pre-processor directives and `#c'
+--
+-- * we lex `#c' as a directive and special case it in the action
+--
+cpp :: CHSLexer
+cpp = directive
+      where
+        directive = 
+	  string "\n#" +> alt ('\t':inlineSet)`star` char '\n'
+	  `lexmeta` 
+	     \(_:_:dir) pos s ->	-- strip off the "\n#"
+	       case dir of
+	         'c':sp:_ | sp `elem` " \t\n" ->	-- #c
+		   (Nothing, retPos (retPos pos), s, Just cLexer)
+                 _                            ->        -- CPP directive
+		   (Just $ Right (CHSTokCPP pos (init dir)), 
+		    retPos (retPos pos), s, Nothing)
+
 -- the binding hook lexer
 --
 bhLexer :: CHSLexer
 bhLexer  =      identOrKW
 	   >||< symbol
 	   >||< strlit
+	   >||< hsverb
 	   >||< whitespace
 	   >||< endOfHook
 	   >||< string "--" +> anyButNL`star` char '\n'	  -- comment
@@ -491,6 +594,21 @@ bhLexer  =      identOrKW
 			  \_ pos s -> (Just $ Right (CHSTokEndHook pos), 
 				       incPos pos 2, s, Just chslexer)
 
+-- the inline-C lexer
+--
+cLexer :: CHSLexer
+cLexer =      inlineC			  -- inline C code
+	 >||< ctrl		          -- control code (preserved)
+         >||< string "\n#endc"		  -- end of inline C code...
+	      `lexmeta`			  -- ...preserve '\n' as control token
+	      \_ pos s -> (Just $ Right (CHSTokCtrl pos '\n'), retPos pos, s, 
+			   Just chslexer)
+	 where
+	   inlineC = alt inlineSet `lexaction` copyVerbatimC
+	   --
+	   copyVerbatimC :: CHSAction 
+	   copyVerbatimC cs pos = Just $ CHSTokC pos cs
+
 -- whitespace
 --
 -- * horizontal and vertical tabs, newlines, and form feeds are filter out by
@@ -512,18 +630,19 @@ identOrKW  =
   where
     idkwtok pos "as"               _    = CHSTokAs      pos
     idkwtok pos "call"             _    = CHSTokCall    pos
+    idkwtok pos "class"            _    = CHSTokClass   pos
     idkwtok pos "context"          _    = CHSTokContext pos
     idkwtok pos "deriving"	   _	= CHSTokDerive  pos
     idkwtok pos "enum"             _    = CHSTokEnum    pos
     idkwtok pos "foreign"	   _	= CHSTokForeign pos
     idkwtok pos "fun"              _    = CHSTokFun     pos
     idkwtok pos "get"              _    = CHSTokGet     pos
-    idkwtok pos "header"           _    = CHSTokHeader  pos
     idkwtok pos "import"           _    = CHSTokImport  pos
     idkwtok pos "lib"              _    = CHSTokLib     pos
     idkwtok pos "newtype"          _    = CHSTokNewtype pos
     idkwtok pos "pointer"          _    = CHSTokPointer pos
     idkwtok pos "prefix"           _    = CHSTokPrefix  pos
+    idkwtok pos "pure"             _    = CHSTokPure    pos
     idkwtok pos "qualified"        _    = CHSTokQualif  pos
     idkwtok pos "set"              _    = CHSTokSet     pos
     idkwtok pos "sizeof"           _    = CHSTokSizeof  pos
@@ -539,10 +658,14 @@ identOrKW  =
 --
 symbol :: CHSLexer
 symbol  =      sym "->" CHSTokArrow
+	  >||< sym "=>" CHSTokDArrow
 	  >||< sym "."  CHSTokDot
 	  >||< sym ","  CHSTokComma
 	  >||< sym "="  CHSTokEqual
+	  >||< sym "-"  CHSTokMinus
 	  >||< sym "*"  CHSTokStar
+	  >||< sym "&"  CHSTokAmp
+	  >||< sym "^"  CHSTokHat
 	  >||< sym "{"  CHSTokLBrace
 	  >||< sym "}"  CHSTokRBrace
 	  >||< sym "("  CHSTokLParen
@@ -553,23 +676,31 @@ symbol  =      sym "->" CHSTokArrow
 -- string
 --
 strlit :: CHSLexer
-strlit  = char '"' +> instr`star` char '"'
+strlit  = char '"' +> (instr >|< char '\\')`star` char '"'
 	  `lexaction` \cs pos -> Just (CHSTokString pos (init . tail $ cs))
 
+-- verbatim code
+--
+hsverb :: CHSLexer
+hsverb  = char '`' +> inhsverb`star` char '\''
+	  `lexaction` \cs pos -> Just (CHSTokHSVerb pos (init . tail $ cs))
+
 
 -- regular expressions
 --
-letter, digit, instr :: Regexp s t
-letter = alt ['a'..'z'] >|< alt ['A'..'Z'] >|< char '_'
-digit  = alt ['0'..'9']
-instr  = alt ([' '..'\127'] \\ ['\"'])
+letter, digit, instr, inchar, inhsverb :: Regexp s t
+letter   = alt ['a'..'z'] >|< alt ['A'..'Z'] >|< char '_'
+digit    = alt ['0'..'9']
+instr    = alt ([' '..'\127'] \\ "\"\\")
+inchar   = alt ([' '..'\127'] \\ "\'")
+inhsverb = alt ([' '..'\127'] \\ "\'")
 
 -- character sets
 --
 anySet, inlineSet, specialSet, commentSpecialSet, ctrlSet :: [Char]
 anySet            = ['\0'..'\255']
 inlineSet         = anySet \\ ctrlSet
-specialSet        = ['{', '-', '"']
+specialSet        = ['{', '-', '"', '\'']
 commentSpecialSet = ['{', '-']
 ctrlSet           = ['\n', '\f', '\r', '\t', '\v']
 
